#pragma once

#define FD_FORCEREAD (1<<15)

#include "winsock2.h"

#pragma warning(disable:4786)

#include <map>
using namespace std;

class CSocketHelperWnd;
class CAsyncSocketExLayer;
class CAsyncSocketEx;

#define DEFAULT_SMT_ITEM_COUNT	8192

typedef enum _E_SMT_ITEM_STATE
{
	eSS_NOT_USED,
	eSS_USED,
	eSS_DEAD
} E_SMT_ITEM_STATE;

typedef struct _SOCKET_MAPPING_TABLE_ITEM
{
	E_SMT_ITEM_STATE eState;
	union
	{
		CAsyncSocketEx* pSocket;
		int nDeadCount;
	};
//	SYSTEMTIME last_state_changed_st;
	
} SOCKET_MAPPING_TABLE_ITEM;


typedef std::map<CAsyncSocketEx *, CAsyncSocketEx*> socketmap_t;


/**
  @ingroup NpSessionBase
  @brief CAsyncSocketEx is a replacement for the MFC class CAsyncSocket.

 <br>
 added by ickar<br>
 modified by ickar<br>
 <br>
 WASAsyncSelect() api  ̿Ͽ single thread  ޽ ̺Ʈ    ̺Ʈ   wrapping  Ŭ.<br>
   ¼ҽ ̺귯 κ Ͽ (:http://www.codeproject.com/internet/casyncsocketex.asp) <br>
<br>
<b>Introduction:</b><br>

CAsyncSocketEx is a replacement for the MFC class CAsyncSocket.
This class was written because CAsyncSocket is not the fastest WinSock
wrapper and it's very hard to add new functionality to CAsyncSocket
derived classes. This class offers the same functionality as CAsyncSocket.
Also, CAsyncSocketEx offers some enhancements which were not possible with
CAsyncSocket without some tricks.<br>
<br>
<b>How do I use it?</b><br>
Basically exactly like CAsyncSocket.
To use CAsyncSocketEx, just replace all occurrences of CAsyncSocket in your
code with CAsyncSocketEx. If you did not enhance CAsyncSocket yourself in
any way, you won't have to change anything else in your code.<br>
<br>
<b>Why is CAsyncSocketEx faster?</b><br>
CAsyncSocketEx is slightly faster when dispatching notification event messages.
First have a look at the way CAsyncSocket works. For each thread that uses
CAsyncSocket, a window is created. CAsyncSocket calls WSAAsyncSelect with
the handle of that window. Until here, CAsyncSocketEx works the same way.
But CAsyncSocket uses only one window message (WM_SOCKET_NOTIFY) for all
sockets within one thread. When the window recieve WM_SOCKET_NOTIFY, wParam
contains the socket handle and the window looks up an CAsyncSocket instance
using a map. CAsyncSocketEx works differently. It's helper window uses a
wide range of different window messages (WM_USER through 0xBFFF) and passes
a different message to WSAAsyncSelect for each socket. When a message in
the specified range is received, CAsyncSocketEx looks up the pointer to a
CAsyncSocketEx instance in an Array using the index of message - WM_USER.
As you can see, CAsyncSocketEx uses the helper window in a more efficient
way, as it don't have to use the slow maps to lookup it's own instance.
Still, speed increase is not very much, but it may be noticeable when using
a lot of sockets at the same time.
Please note that the changes do not affect the raw data throughput rate,
CAsyncSocketEx only dispatches the notification messages faster.<br>
<br>
<b>What else does CAsyncSocketEx offer?</b><br>
CAsyncSocketEx offers a flexible layer system. One example is the proxy layer.
Just create an instance of the proxy layer, configure it and add it to the layer
chain of your CAsyncSocketEx instance. After that, you can connect through
proxies.
Benefit: You don't have to change much to use the layer system.
Another layer that is currently in development is the SSL layer to establish
SSL encrypted connections.<br>
<br>
<b>License</b><br>
Feel free to use this class, as long as you don't claim that you wrote it
and this copyright notice stays intact in the source files.
If you use this class in commercial applications, please send a short message
to tim.kosse@gmx.de


  <b>߰</b><br>
  2003-10-20 翵(Yang Young Seung); ickar@nate.com<br>
 
   CAsyncSocketEx  ɰ ׸  ִ.
  1  ؼ Connect  ϰ, closesocket()    ̻  ʴ´ٸ  ߻ ʴ´.
  , 1  , Connect ϰ Send ϰ ϴ  ,   closesocket() ϰ ٽ ٸ
  Connect ؼ Ѵٰ  . ̷   ſ ڿ ̴.<br>
 
     closesocket()  ٽ connect ϱ ؼ socket() api Լ ؼ ο    ̴.
  ׷ ̶     ڵ Ѵ. ׷    ɼ   ,,
  Ŭ Ʈ ù°   ϴ  ް, closesocket() Լ θ,     
  close()   ִ. ׸  close Ǿ Ŭ̾Ʈ closesocket() Լ θ Client  OS Ŀ  
  ߴٸ, ޽ ť FD_CLOSE ̺Ʈ Post Ǿ ̴. ̷  Ŀ closesocket() Լ Ҹٰ ص ̹
  ޽ ť  ִ FD_CLOSE ̺Ʈ ޽  ʴ´. ׷ٸ, ״ Connect () Լ θĿ..  ̺Ʈ 
  OnClose() ̺Ʈ  ȴ. ̰츦  ϱ ؼ,, MFC  CAsyncSocket class  Ͽ,,
  KillSocket() ̶ static Method  ߰ Ͽ. 
     ̹ ޽ ť  Ǿ ִ  Ƽ  ޽  ſ  ,
  WM_SOCKETEX_DEAD  ޽ closesocket()  ȣϴ  ϰ, ̸޽  
  õ socket notify message   ϴ°̴.
 
 @author Tim Kosse (Tim.Kosse@gmx.de)

 @version 1.2 (2003-03-28)
*/

class CAsyncSocketEx
{

//////////////////////////////////////////////////////////////////////////////
//Functions that imitate CAsyncSocket
//////////////////////////////////////////////////////////////////////////////

	bool m_bIsTcpListenSocket;
protected:
	int m_nSMT_Index;
private:

	CAsyncSocketEx(const CAsyncSocketEx& rSrc);    // no implementation
	void operator=(const CAsyncSocketEx& rSrc);  // no implementation

//Construction
public:
	CAsyncSocketEx();
	BOOL Create(UINT nSocketPort = 0, int nSocketType=SOCK_STREAM,
		long lEvent = FD_READ | FD_WRITE | FD_OOB | FD_ACCEPT |	FD_CONNECT | FD_CLOSE,
		LPCTSTR lpszSocketAddress = NULL);

	bool m_bNotifyClosed;

	BOOL TerminateAsyncSelect();

	void Delete()
	{
		delete this;
	}
//Attributes
public:
	bool m_bRequestConnect;
	SOCKET m_hSocket;
	
	operator SOCKET() const;
	BOOL Attach(SOCKET hSocket, long lEvent = FD_READ | FD_WRITE | FD_OOB | FD_ACCEPT | FD_CONNECT | FD_CLOSE );
	SOCKET Detach();

#ifdef _AFX
	BOOL GetPeerName( CString& rPeerAddress, UINT& rPeerPort );
#endif
	BOOL GetPeerName( SOCKADDR* lpSockAddr, int* lpSockAddrLen );

#ifdef _AFX
	BOOL GetSockName( CString& rSocketAddress, UINT& rSocketPort );
#endif
	BOOL GetSockName( SOCKADDR* lpSockAddr, int* lpSockAddrLen );

	BOOL GetSockOpt(int nOptionName, void* lpOptionValue, 
		int* lpOptionLen, int nLevel = SOL_SOCKET);
	BOOL SetSockOpt(int nOptionName, const void* lpOptionValue, 
		int nOptionLen, int nLevel = SOL_SOCKET);

	static int PASCAL GetLastError();

//Operations
public:

	virtual BOOL Accept( CAsyncSocketEx& rConnectedSocket, 
		SOCKADDR* lpSockAddr = NULL, int* lpSockAddrLen = NULL );

	BOOL Bind(UINT nSocketPort, LPCTSTR lpszSocketAddress = NULL);
	BOOL Bind(const SOCKADDR* lpSockAddr, int nSockAddrLen);

	virtual void Close();

	static BOOL InitSocket();

	virtual BOOL Connect(LPCTSTR lpszHostAddress, UINT nHostPort, int nTimeOut = -1 );
	virtual BOOL Connect( const SOCKADDR* lpSockAddr, int nSockAddrLen, int nTimeOut = -1 );

	BOOL IOCtl( long lCommand, DWORD* lpArgument );

	BOOL Listen( int nConnectionBacklog = 5 );


	virtual int Receive(void* lpBuf, int nBufLen, int nFlags = 0);

#ifdef _AFX
	int ReceiveFrom(void* lpBuf, int nBufLen,
		CString& rSocketAddress, UINT& rSocketPort, int nFlags = 0);
#endif
	int ReceiveFrom(void* lpBuf, int nBufLen,
		SOCKADDR* lpSockAddr, int* lpSockAddrLen, int nFlags = 0);

	enum { receives = 0, sends = 1, both = 2 };
	BOOL ShutDown( int nHow = sends );

	virtual int Send(const void* lpBuf, int nBufLen, int nFlags = 0);

#ifdef _AFX
	int SendTo(const void* lpBuf, int nBufLen,
		UINT nHostPort, LPCTSTR lpszHostAddress = NULL, int nFlags = 0);
#endif
	int SendTo(const void* lpBuf, int nBufLen,
		const SOCKADDR* lpSockAddr, int nSockAddrLen, int nFlags = 0);
	
	BOOL AsyncSelect(long lEvent = FD_READ | FD_WRITE | FD_OOB | FD_ACCEPT | FD_CONNECT | FD_CLOSE);
	
	int SendWSA(LPWSABUF lpBuffers,
		DWORD dwBufferCount,
		LPDWORD lpNumberOfBytesSent,
		DWORD dwFlags,
		LPWSAOVERLAPPED lpOverlapped,
		LPWSAOVERLAPPED_COMPLETION_ROUTINE lpCompletionRoutine);
	
	// Overridable callbacks
protected:
	virtual void OnReceive(int nErrorCode, int nBytes);
	virtual void OnSend(int nErrorCode);
	virtual void OnOutOfBandData(int nErrorCode);
	virtual void OnAccept(int nErrorCode);
	virtual void OnConnect(int nErrorCode);
	virtual void OnClose(int nErrorCode);
	
	// Implementation
public:
	virtual ~CAsyncSocketEx();
	
	static SOCKET_MAPPING_TABLE_ITEM * ms_pSocketMappingTable;
	static int ms_nSocketMappingTableSize;
	static int ms_nSocketMappingTable_LastIndex;
	static int ms_nSocketMappingTable_UsedItemCount;
	static int ms_nSocketMappingTable_DeadItemCount;
	static int PASCAL sFind_SMT_NextAvailableIndex();

	static CSocketHelperWnd *ms_pHelperWindow;

	static bool PASCAL sAttachHandle(CAsyncSocketEx* pSocket);
	static void PASCAL sDetachHandle(int nSMT_Index);

	static void PASCAL sAddSocketDeadMark(int nSMT_Index);
	static void PASCAL sRemoveSocketDeadMark(int nSMT_Index);

	static void PASCAL sKillSocket(int nSMT_Index, CAsyncSocketEx* pSocket);
	static void PASCAL sDoCallBack(int nSMT_Index, WPARAM wParam, LPARAM lParam);
	
	BOOL Socket(int nSocketType=SOCK_STREAM, long lEvent =
		FD_READ | FD_WRITE | FD_OOB | FD_ACCEPT | FD_CONNECT | FD_CLOSE,
		int nProtocolType = 0, int nAddressFormat = PF_INET);
	
protected:
	friend class CSocketHelperWnd;

/////////////////////////////////////////////////////////////////////////////////////
// Additional functions
/////////////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////////////
// support Socket Helper Windows
protected:

	//Returns the handle of the helper window
	HWND GetHelperWindowHandle();


/////////////////////////////////////////////////////////////////////////////////////
//AsyncGetHostByName
public:
	char *m_pAsyncGetHostByNameBuffer; //Buffer for hostend structure
	HANDLE m_hAsyncGetHostByNameHandle; //TaskHandle
	char * m_pszAsyncGetHostByNameDomain;
	int m_nAsyncGetHostByNamePort; //Port to connect to


/////////////////////////////////////////////////////////////////////////////////////
// Support CAsyncSocketExLayer
public:
	friend CAsyncSocketExLayer;

	//Attaches a new layer to the socket.
	BOOL AddLayer(CAsyncSocketExLayer *pLayer);

	//Resets layer chain.
	void RemoveAllLayers();

	//Returns the handle of the socket.
	SOCKET GetSocketHandle();

protected:
	//Trigers an event on the socket
	// Any combination of FD_READ, FD_WRITE, FD_CLOSE, FD_ACCEPT, FD_CONNECT and FD_FORCEREAD is valid for lEvent.
	BOOL TriggerEvent(short lEvent, short error);

protected:

	int m_nConnectionTimeout;

	//If using layers, only the events specified with m_lEvent will send to the event handlers.
	long m_lEvent;

	//Layer chain
	CAsyncSocketExLayer *m_pFirstLayer;
	CAsyncSocketExLayer *m_pLastLayer;

	//Called by the layers to notify application of some events
	virtual int OnLayerCallback(const CAsyncSocketExLayer *pLayer, int nType, int nParam1, int nParam2);

	/////////////////////////////////////////////////////////////////////////
	// 2003-07-26 翵 ߰
	// FD_WRITE ̺Ʈ  ޱ ؼ.. .. WSAAsyncSelect()  θ̴.
	void CheckOnSendEvent();
	void UnCheckOnSendEvent() { }

	static void sDumpMappingTable();
};

inline
SOCKET CAsyncSocketEx::GetSocketHandle()
{
	return m_hSocket;
}

inline
BOOL CAsyncSocketEx::IOCtl( long lCommand, DWORD* lpArgument )
{
	return ioctlsocket(m_hSocket, lCommand, lpArgument) != SOCKET_ERROR;
}

// static
inline
int CAsyncSocketEx::GetLastError()
{
	return WSAGetLastError();
}



#define LAYERCALLBACK_STATECHANGE		0
#define LAYERCALLBACK_LAYERSPECIFIC		1



/**
  @ingroup NpSessionBase
  @brief κ ̺Ʈ ޽   ü Ŭ.
*/
class CSocketHelperWnd
{
	static TCHAR ms_szSocketWndClassName[256];
	static HINSTANCE ms_hInstance;

	// Connection Timeout Feature
	static socketmap_t ms_mapConnectionSocketList;
	static CRITICAL_SECTION	ms_hCriticalSection;
	static BOOL m_bConnectionCheckTimerActivate;

public:
	static void sRegisterClass(HINSTANCE hInstance);
	static void sUnregisterClass();

	static void sAddConnectionSocketToMap(CAsyncSocketEx *pSocket);
	static void sRemoveConnectionSocketFromMap(CAsyncSocketEx *pSocket);

	static void sStartConnectionTimer();
	static void sKillConnectionTimer();

	CSocketHelperWnd();
	virtual ~CSocketHelperWnd();

	//Processes event notifications sent by the sockets or the layers
	static LRESULT CALLBACK WindowProc( HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam );

	HWND CSocketHelperWnd::GetHwnd()
	{
		return m_hWnd;
	}

	map<HANDLE, CAsyncSocketEx *> m_mapAsyncGetHostByName;

private:
	HWND m_hWnd;
};

